

open Graph
open Graph.Dot_ast
open Data
open Event
open Topology
    
type process = {
  name: string;
  actions: (string * bool * bool) list; (* (action name, enabled, active) *)
  vars: (string * Data.v) list (* pid local vars*)
}

let (is_parent: string -> int -> Event.t -> bool) =
  fun a i e ->
    (* XXX marche ssi une variable s'appelle par!
       je devrais au moins generaliser avec l'existence
       d'une variable de type parent (et encore)
    *)
      match  List.assoc_opt (a^"_par") e.data with
      | None -> false
      | Some (I j) -> j > -1 && i = j
      | _ -> false

let (get_processes : string -> Event.t -> process list) =
  fun f e -> 
    if e.kind <> Ltop then (
      print_string "print_dot should be called from Ltop event\n";
      failwith "exit print_dot"
    );
    let l = List.map (fun (x,v) -> Str.split (Str.regexp "_") x, v) e.data in
    let rec sortv (enab, other) (x,v) =
      match x with 
      | ["Enab";pid;a] -> (pid,a,v)::enab, other
      | pid::tail -> enab, (pid,(String.concat "_" tail),v)::other
      | [] -> assert false
    in
    let enab, other = List.fold_left sortv ([],[]) l in
    let rec (build_pidl: process list -> (string * string * Data.v) list ->
             (string * string * Data.v) list -> process list) =
      fun pidl enab other ->
        match enab with
        | [] -> pidl 
        | (pid, _, _)::_ ->
          let enab_pid_list, enab = List.partition (fun (pid0,_,_) -> pid=pid0) enab in
          let other_pid, other = List.partition (fun (pid0,_,_) -> pid=pid0) other in
          let acti_pid, vars_pid =
            List.partition
              (fun (_,n,_) -> List.exists (fun (_,n2,_) -> n2=n) enab_pid_list)
              other_pid
          in
          let get_actions (_, n, enabv) =
            let (_,_,activ) = List.find (fun (_,n0, _) -> n=n0) acti_pid in
            (n, enabv = Data.B true, activ = B true)
          in
          let pid = {
            name = pid;
            actions = List.map get_actions enab_pid_list;
            vars = List.map (fun (_,n,v) -> n,v) vars_pid;
          }
          in
          build_pidl (pid::pidl) enab other
    in
    let pidl = build_pidl [] enab other in
    List.rev pidl
  

(* Compute a dot from the content of e.data  *)
let to_pdf engine g f e =
  let nodes = g.nodes in
  let (pidl : process list) = get_processes f e in
  let oc = open_out ("sasa-"^f) in
  let nodes_decl =
    String.concat "\n"
      (List.map
         (fun pid ->
            let color =
              if List.exists (fun (_,_,a) -> a) pid.actions then
                "fillcolor=gold,style=filled,"
              else
              if List.exists (fun (_,e,_) -> e) pid.actions then
                "fillcolor=green,style=filled,"
              else ""
            in
            let enabled = String.concat ","
                (List.map
                   (fun (n,_,_) -> n)
                   (List.filter (fun (_,e,_) -> e) pid.actions))
            in
            let enabled = if enabled = "" then "" else (enabled^"|") in
            let loc = String.concat "|"
                (List.map (fun (n,v) ->
                     Printf.sprintf "%s=%s" n (Data.val_to_string string_of_float v))
                    pid.vars
                )
            in
            Printf.sprintf " %s [%slabel=\"%s|{%s%s}\"] "
              pid.name color pid.name enabled loc
         )
         pidl
      )
  in
  let trans = 
    List.flatten
      (List.map
         (fun n ->
            let l = g.succ n.id in
            List.mapi (fun i t ->
                if is_parent n.id i e then
                  Printf.sprintf "%s -> %s" n.id t 
                else if n.id < t then
                  Printf.sprintf "%s -- %s" n.id t
                else
                  Printf.sprintf "%s -- %s" t n.id
              )
              l
         )
         nodes
      )
  in
  let trans = List.sort_uniq compare trans in
  let is_directed str =
    try
      ignore (Str.search_forward (Str.regexp "->") str  0);
      true
    with Not_found -> false
  in
  let trans_dir,trans_undir = List.partition is_directed trans in
  let trans_dir_str = String.concat "\n" trans_dir in
  let trans_undir_str = String.concat "\n" trans_undir in
  let trans_undir_str = Str.global_replace (Str.regexp "--") "->" trans_undir_str in
  let trans_str =
    (*     if trans_dir_str = "" then trans_undir_str else *)
    Printf.sprintf "subgraph dir {\n\t%s} 
subgraph undir {\n\t edge [dir=none]\n%s} " trans_dir_str trans_undir_str
  in
  Printf.fprintf oc
    "digraph %s {\nlabel=\"%s step %d\"\nnode [shape=record];\n%s\n%s\n}\n"
    "g" f e.step 
    nodes_decl trans_str;
  flush oc;
  close_out oc;
  Sys.command (Printf.sprintf "%s -Tpdf sasa-%s -o sasa-%s.pdf" engine f f)
;;

let dot = to_pdf "dot"  ;;
let neato = to_pdf "neato"  ;;
let twopi = to_pdf "twopi"  ;;
let circo = to_pdf "circo"  ;;
let fdp = to_pdf "fdp"  ;;
let sfdp = to_pdf "sfdp"  ;;
let patchwork = to_pdf "patchwork" ;;
let osage = to_pdf "osage"  ;;
(*
  From the dot man:
       dot - filter for drawing directed graphs
       neato - filter for drawing undirected graphs
       twopi - filter for radial layouts of graphs
       circo - filter for circular layout of graphs
       fdp - filter for drawing undirected graphs
       sfdp - filter for drawing large undirected graphs
       patchwork - filter for squarified tree maps
       osage - filter for array-based layouts
  *)

(***********************************************************************)

let get_removable pl =
  let pl = List.filter
      (fun p ->
         (List.exists (fun (_,_,acti) -> acti) p.actions) ||
         (List.for_all (fun (_,enab,_) -> (not enab)) p.actions)
      )
      pl
  in
  List.map (fun p -> p.name) pl
 
let next_round nodes f e = 
  let (pl : process list) = get_processes f e in
  let p_with_enable_action =
    List.filter
      (fun p -> List.exists (fun (_,enab,acti) -> enab&&not(acti)) p.actions)
      pl
  in
  let pidl = List.map (fun p -> p.name) p_with_enable_action in
  Printf.printf "The processes to check : %s\n" (String.concat "," pidl);
  flush stdout;
  let rec go cpidl e =
    let e = step e in
    let pl = get_processes f e in
    let removable = get_removable pl in
    Printf.printf "Current process: %s\n" (String.concat "," cpidl);
    Printf.printf "Removable process: %s\n" (String.concat "," removable);
    flush stdout;
    let cpidl = List.filter (fun pid -> not (List.mem pid removable)) cpidl in
    if cpidl = [] then e else go cpidl e
  in
  go pidl e
    
(***********************************************************************)

let _ = print_string "
===> Use the read fonction to load the dot file
===> Use the dot function at Ltop event to generated .dot.pdf files

You migth want to add something along those lines at the end of your rdbg-session.ml file
(the name of the dot file and the path to the dot.ml file migth need to be adapted
to your context tough):
;;
#require \"ocamlgraph\";;
#mod_use \"../../lib/algo/algo.ml\";;
#use \"../../lib/sasacore/topology.ml\";;
#use \"../rdbg-utils/dot.ml\";;
let dotfile = \"g.dot\";;
let p = Topology.read dotfile;;
let d () = dot (p.nodes, dotfile) !e;;
let sd () = s();d();;
let nr () = e:=next_round p.nodes dotfile !e; d();;

let _ = n (); d (); Sys.command (\"zathura sasa-g.dot.pdf&\")

"
